# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
from typing import Any, Dict, List, Optional, Protocol

from swh.core.statsd import Statsd
from swh.journal.client import get_journal_client
from swh.journal.serializers import kafka_to_value, value_to_kafka
from swh.model.model import Content
from swh.model.swhids import ObjectType
from swh.objstorage.exc import ObjCorruptedError, ObjNotFoundError
from swh.objstorage.interface import ObjStorageInterface, objid_from_dict
from swh.storage.interface import StorageInterface

from .base_checker import BaseChecker, BasePartitionChecker
from .db import ConfigEntry, Datastore, ScrubberDb

logger = logging.getLogger(__name__)


def get_objstorage_datastore(objstorage_config):
    objstorage_config = dict(objstorage_config)
    return Datastore(
        package="objstorage",
        cls=objstorage_config.pop("cls"),
        instance=json.dumps(objstorage_config),
    )


class ObjectStorageCheckerProtocol(Protocol):
    db: ScrubberDb
    objstorage: ObjStorageInterface

    @property
    def config(self) -> ConfigEntry: ...

    @property
    def statsd(self) -> Statsd: ...


class ContentCheckerMixin(ObjectStorageCheckerProtocol):
    """Mixin class implementing content checks used by object storage checkers."""

    def check_content(self, content: Content) -> None:
        """Checks if a content exists in an object storage (if ``check_references`` is set to
        :const:`True` in checker config) or if a content is corrupted in an object storage (if
        ``check_hashes`` is set to :const:`True` in checker config).
        """

        content_hashes = objid_from_dict(content.hashes())
        try:
            self.objstorage.check(content_hashes)
        except ObjNotFoundError:
            if self.config.check_references:
                self.statsd.increment("missing_object_total")
                self.db.missing_object_add(
                    id=content.swhid(), reference_ids={}, config=self.config
                )
        except ObjCorruptedError:
            if self.config.check_hashes:
                self.statsd.increment("hash_mismatch_total")
                self.db.corrupt_object_add(
                    id=content.swhid(),
                    config=self.config,
                    serialized_object=value_to_kafka(content.to_dict()),
                )


class ObjectStorageCheckerFromStoragePartition(
    BasePartitionChecker, ContentCheckerMixin
):
    """A partition based checker to detect missing and corrupted contents in an object storage.

    It iterates on content objects referenced in a storage instance, check they are available
    in a given object storage instance (if ``check_references`` is set to :const:`True` in
    checker config) then retrieve their bytes from it in order to recompute checksums and detect
    corruptions (if ``check_hashes`` is set to :const:`True` in checker config)."""

    def __init__(
        self,
        db: ScrubberDb,
        config_id: int,
        storage: StorageInterface,
        objstorage: Optional[ObjStorageInterface] = None,
        limit: int = 0,
    ):
        super().__init__(db=db, config_id=config_id, limit=limit)
        self.storage = storage
        self.objstorage = (
            objstorage if objstorage is not None else getattr(storage, "objstorage")
        )

        object_type = self.object_type.name.lower()

        if object_type != "content":
            raise ValueError(
                "ObjectStorageCheckerFromStoragePartition can only check objects of type "
                f"content, checking objects of type {object_type} is not supported."
            )

        if self.objstorage is None:
            raise ValueError(
                "An object storage must be provided to ObjectStorageCheckerFromStoragePartition."  # noqa
            )

    def check_partition(self, object_type: ObjectType, partition_id: int) -> None:
        page_token = None
        while True:
            page = self.storage.content_get_partition(
                partition_id=partition_id,
                nb_partitions=self.nb_partitions,
                page_token=page_token,
            )
            contents = page.results

            with self.statsd.timed(
                "batch_duration_seconds", tags={"operation": "check_hashes"}
            ):
                logger.debug("Checking %s content object hashes", len(contents))
                for content in contents:
                    self.check_content(content)

            page_token = page.next_page_token
            if page_token is None:
                break


class ObjectStorageCheckerFromJournal(BaseChecker, ContentCheckerMixin):
    """A journal based checker to detect missing and corrupted contents in an object storage.

    It iterates on content objects referenced in a kafka topic, check they are available
    in a given object storage instance then retrieve their bytes from it in order to
    recompute checksums and detect corruptions."""

    def __init__(
        self,
        db: ScrubberDb,
        config_id: int,
        journal_client_config: Dict[str, Any],
        objstorage: ObjStorageInterface,
    ):
        super().__init__(db=db, config_id=config_id)
        self.objstorage = objstorage

        object_type = self.object_type.name.lower()

        if object_type != "content":
            raise ValueError(
                "ObjectStorageCheckerFromJournal can only check objects of type content,"
                f"checking objects of type {object_type} is not supported."
            )

        self.journal_client_config = journal_client_config.copy()
        if "object_types" in self.journal_client_config:
            raise ValueError(
                "The journal_client configuration entry should not define the "
                "object_types field; this is handled by the scrubber configuration entry"
            )
        self.journal_client_config["object_types"] = [object_type]
        self.journal_client = get_journal_client(
            **self.journal_client_config,
            # Remove default deserializer; so process_kafka_values() gets the message
            # verbatim so it can archive it with as few modifications a possible.
            value_deserializer=lambda obj_type, msg: msg,
        )

    def run(self) -> None:
        self.journal_client.process(self.process_kafka_messages)

    def process_kafka_messages(self, all_messages: Dict[str, List[bytes]]):
        for message in all_messages["content"]:
            self.check_content(Content.from_dict(kafka_to_value(message)))



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from .objstorage_checker_tests import *  # noqa

# Use postgreql storage and an objstorage with pathslicing backend to run
# the tests


def swh_objstorage_config(tmpdir):
    return {
        "cls": "pathslicing",
        "root": str(tmpdir),
        "slicing": "0:2/2:4/4:6",
        "compression": "gzip",
    }



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Reads all objects in a swh-storage instance and recomputes their checksums."""

import collections
import contextlib
import json
import logging
from typing import Iterable, Optional, Tuple, Union

from swh.journal.serializers import value_to_kafka
from swh.model import swhids
from swh.model.model import (
    Content,
    Directory,
    ObjectType,
    Release,
    Revision,
    Snapshot,
    SnapshotTargetType,
)
from swh.storage.algos.directory import (
    directory_get_many_with_possibly_duplicated_entries,
)
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.storage.cassandra.storage import CassandraStorage
from swh.storage.interface import StorageInterface
from swh.storage.postgresql.storage import Storage as PostgresqlStorage

from .base_checker import BasePartitionChecker
from .db import Datastore, ScrubberDb

logger = logging.getLogger(__name__)

ScrubbableObject = Union[Revision, Release, Snapshot, Directory, Content]


@contextlib.contextmanager
def postgresql_storage_db(storage):
    db = storage.get_db()
    try:
        yield db
    finally:
        storage.put_db(db)


def _get_inclusive_range_swhids(
    inclusive_range_start: Optional[bytes],
    exclusive_range_end: Optional[bytes],
    object_type: swhids.ObjectType,
) -> Tuple[swhids.CoreSWHID, swhids.CoreSWHID]:
    r"""
    Given a ``[range_start, range_end)`` right-open interval of id prefixes
    and an object type (as returned by :const:`swh.storage.backfill.RANGE_GENERATORS`),
    returns a ``[range_start_swhid, range_end_swhid]`` closed interval of SWHIDs
    suitable for the scrubber database.

    >>> _get_inclusive_range_swhids(b"\x42", None, swhids.ObjectType.SNAPSHOT)
    (CoreSWHID.from_string('swh:1:snp:4200000000000000000000000000000000000000'), CoreSWHID.from_string('swh:1:snp:ffffffffffffffffffffffffffffffffffffffff'))

    >>> _get_inclusive_range_swhids(b"\x00", b"\x12\x34", swhids.ObjectType.REVISION)
    (CoreSWHID.from_string('swh:1:rev:0000000000000000000000000000000000000000'), CoreSWHID.from_string('swh:1:rev:1233ffffffffffffffffffffffffffffffffffff'))

    """  # noqa
    range_start_swhid = swhids.CoreSWHID(
        object_type=object_type,
        object_id=(inclusive_range_start or b"").ljust(20, b"\00"),
    )
    if exclusive_range_end is None:
        inclusive_range_end = b"\xff" * 20
    else:
        # convert "1230000000..." to "122fffffff..."
        inclusive_range_end = (
            int.from_bytes(exclusive_range_end.ljust(20, b"\x00"), "big") - 1
        ).to_bytes(20, "big")
    range_end_swhid = swhids.CoreSWHID(
        object_type=object_type,
        object_id=inclusive_range_end,
    )

    return (range_start_swhid, range_end_swhid)


def get_datastore(storage) -> Datastore:
    if isinstance(storage, PostgresqlStorage):
        with postgresql_storage_db(storage) as db:
            datastore = Datastore(
                package="storage",
                cls="postgresql",
                instance=db.conn.dsn,
            )
    elif isinstance(storage, CassandraStorage):
        datastore = Datastore(
            package="storage",
            cls="cassandra",
            instance=json.dumps(
                {
                    "keyspace": storage.keyspace,
                    "hosts": storage.hosts,
                    "port": storage.port,
                }
            ),
        )
    else:
        raise NotImplementedError(f"StorageChecker(storage={storage!r}).datastore()")
    return datastore


class StorageChecker(BasePartitionChecker):
    """Reads a chunk of a swh-storage database, recomputes checksums, and
    reports errors in a separate database."""

    def __init__(
        self, db: ScrubberDb, config_id: int, storage: StorageInterface, limit: int = 0
    ):
        super().__init__(db=db, config_id=config_id, limit=limit)
        self.storage = storage

    def check_partition(
        self, object_type: swhids.ObjectType, partition_id: int
    ) -> None:
        page_token = None
        while True:
            if object_type in (swhids.ObjectType.RELEASE, swhids.ObjectType.REVISION):
                method = getattr(
                    self.storage, f"{self.object_type.name.lower()}_get_partition"
                )
                page = method(partition_id, self.nb_partitions, page_token=page_token)
                objects = page.results
            elif object_type == swhids.ObjectType.DIRECTORY:
                page = self.storage.directory_get_id_partition(
                    partition_id, self.nb_partitions, page_token=page_token
                )
                directory_ids = page.results
                objects = []
                for dir_id, item in zip(
                    directory_ids,
                    directory_get_many_with_possibly_duplicated_entries(
                        self.storage, directory_ids
                    ),
                ):
                    assert item is not None, f"Directory {dir_id.hex()} disappeared"
                    (has_duplicate_entries, object_) = item
                    if has_duplicate_entries:
                        self.statsd.increment("duplicate_directory_entries_total")
                        self.db.corrupt_object_add(
                            object_.swhid(),
                            self.config,
                            value_to_kafka(object_.to_dict()),
                        )
                    objects.append(object_)
            elif object_type == swhids.ObjectType.SNAPSHOT:
                page = self.storage.snapshot_get_id_partition(
                    partition_id, self.nb_partitions, page_token=page_token
                )
                snapshot_ids = page.results
                objects = [
                    snapshot_get_all_branches(self.storage, snapshot_id)
                    for snapshot_id in snapshot_ids
                ]
            else:
                assert False, f"Unexpected object type: {object_type}"

            if self.check_hashes:
                with self.statsd.timed(
                    "batch_duration_seconds", tags={"operation": "check_hashes"}
                ):
                    logger.debug(
                        "Checking %s %s object hashes", len(objects), object_type
                    )
                    self.check_object_hashes(objects)
            if self.check_references:
                with self.statsd.timed(
                    "batch_duration_seconds", tags={"operation": "check_references"}
                ):
                    logger.debug(
                        "Checking %s %s object references", len(objects), object_type
                    )
                    self.check_object_references(objects)

            page_token = page.next_page_token
            if page_token is None:
                break

    def check_object_hashes(self, objects: Iterable[ScrubbableObject]):
        """Recomputes hashes, and reports mismatches."""
        count = 0
        for object_ in objects:
            if isinstance(object_, Content):
                # TODO
                continue
            real_id = object_.compute_hash()
            count += 1
            if object_.id != real_id:
                self.statsd.increment("hash_mismatch_total")
                self.db.corrupt_object_add(
                    object_.swhid(),
                    self.config,
                    value_to_kafka(object_.to_dict()),
                )
        if count:
            self.statsd.increment("objects_hashed_total", count)

    def check_object_references(self, objects: Iterable[ScrubbableObject]):
        """Check all objects references by these objects exist."""
        cnt_references = collections.defaultdict(set)
        dir_references = collections.defaultdict(set)
        rev_references = collections.defaultdict(set)
        rel_references = collections.defaultdict(set)
        snp_references = collections.defaultdict(set)

        for object_ in objects:
            swhid = object_.swhid()

            if isinstance(object_, Content):
                pass
            elif isinstance(object_, Directory):
                for entry in object_.entries:
                    if entry.type == "file":
                        cnt_references[entry.target].add(swhid)
                    elif entry.type == "dir":
                        dir_references[entry.target].add(swhid)
                    elif entry.type == "rev":
                        # dir->rev holes are not considered a problem because they
                        # happen whenever git submodules point to repositories that
                        # were not loaded yet; ignore them
                        pass
                    else:
                        assert False, entry
            elif isinstance(object_, Revision):
                dir_references[object_.directory].add(swhid)
                for parent in object_.parents:
                    rev_references[parent].add(swhid)
            elif isinstance(object_, Release):
                if object_.target is None:
                    pass
                elif object_.target_type == ObjectType.CONTENT:
                    cnt_references[object_.target].add(swhid)
                elif object_.target_type == ObjectType.DIRECTORY:
                    dir_references[object_.target].add(swhid)
                elif object_.target_type == ObjectType.REVISION:
                    rev_references[object_.target].add(swhid)
                elif object_.target_type == ObjectType.RELEASE:
                    rel_references[object_.target].add(swhid)
                else:
                    assert False, object_
            elif isinstance(object_, Snapshot):
                for branch in object_.branches.values():
                    if branch is None:
                        pass
                    elif branch.target_type == SnapshotTargetType.ALIAS:
                        pass
                    elif branch.target_type == SnapshotTargetType.CONTENT:
                        cnt_references[branch.target].add(swhid)
                    elif branch.target_type == SnapshotTargetType.DIRECTORY:
                        dir_references[branch.target].add(swhid)
                    elif branch.target_type == SnapshotTargetType.REVISION:
                        rev_references[branch.target].add(swhid)
                    elif branch.target_type == SnapshotTargetType.RELEASE:
                        rel_references[branch.target].add(swhid)
                    elif branch.target_type == SnapshotTargetType.SNAPSHOT:
                        snp_references[branch.target].add(swhid)
                    else:
                        assert False, (str(object_.swhid()), branch)
            else:
                assert False, object_.swhid()

        missing_cnts = set(
            self.storage.content_missing_per_sha1_git(list(cnt_references))
        )
        missing_dirs = set(self.storage.directory_missing(list(dir_references)))
        missing_revs = set(self.storage.revision_missing(list(rev_references)))
        missing_rels = set(self.storage.release_missing(list(rel_references)))
        missing_snps = set(self.storage.snapshot_missing(list(snp_references)))

        self.statsd.increment(
            "missing_object_total",
            len(missing_cnts),
            tags={"target_object_type": "content"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_dirs),
            tags={"target_object_type": "directory"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_revs),
            tags={"target_object_type": "revision"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_rels),
            tags={"target_object_type": "release"},
        )
        self.statsd.increment(
            "missing_object_total",
            len(missing_snps),
            tags={"target_object_type": "snapshot"},
        )

        for missing_id in missing_cnts:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.CONTENT, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, cnt_references[missing_id], self.config
            )

        for missing_id in missing_dirs:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.DIRECTORY, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, dir_references[missing_id], self.config
            )

        for missing_id in missing_revs:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.REVISION, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, rev_references[missing_id], self.config
            )

        for missing_id in missing_rels:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.RELEASE, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, rel_references[missing_id], self.config
            )

        for missing_id in missing_snps:
            missing_swhid = swhids.CoreSWHID(
                object_type=swhids.ObjectType.SNAPSHOT, object_id=missing_id
            )
            self.db.missing_object_add(
                missing_swhid, snp_references[missing_id], self.config
            )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from abc import ABC, abstractmethod
from itertools import count, islice
import logging
from typing import Any, Dict, Iterable, Optional

import psycopg2
import tenacity

from swh.core.statsd import Statsd
from swh.model import swhids

from .db import ConfigEntry, Datastore, ScrubberDb

logger = logging.getLogger(__name__)


class BaseChecker(ABC):
    """Base Checker class wrapping common features."""

    def __init__(
        self,
        db: ScrubberDb,
        config_id: int,
    ):
        self.db = db
        self.config_id = config_id
        self._config: Optional[ConfigEntry] = None
        self._statsd: Optional[Statsd] = None
        self.statsd_constant_tags: Dict[str, Any] = {
            "object_type": self.object_type.name.lower(),
            "datastore_package": self.datastore.package,
            "datastore_cls": self.datastore.cls,
            "datastore_instance": self.datastore.instance,
        }

    @property
    def config(self) -> ConfigEntry:
        """Returns a :class:`ConfigEntry` instance containing checker configuration."""
        if self._config is None:
            self._config = self.db.config_get(self.config_id)

        assert self._config is not None
        return self._config

    @property
    def datastore(self) -> Datastore:
        """Returns a :class:`Datastore` instance representing the source of data
        being checked."""
        return self.config.datastore

    @property
    def statsd(self) -> Statsd:
        """Returns a :class:`Statsd` instance to send statsd metrics."""
        if self._statsd is None:
            self._statsd = Statsd(
                namespace="swh_scrubber",
                constant_tags=self.statsd_constant_tags,
            )
        return self._statsd

    @property
    def object_type(self) -> swhids.ObjectType:
        """Returns the type of object being checked."""
        return self.config.object_type

    @property
    def check_hashes(self) -> bool:
        return self.config.check_hashes

    @property
    def check_references(self) -> bool:
        return self.config.check_references

    @abstractmethod
    def run(self) -> None:
        """Run the checker processing, derived classes must implement this method."""
        pass


class BasePartitionChecker(BaseChecker):
    """Base class for checkers processing partition of objects."""

    def __init__(
        self,
        db: ScrubberDb,
        config_id: int,
        limit: int = 0,
    ):
        super().__init__(db=db, config_id=config_id)
        self.limit = limit
        self.statsd_constant_tags["nb_partitions"] = self.nb_partitions

    @property
    def nb_partitions(self) -> int:
        """Returns the number of partitions set in configuration."""
        return self.config.nb_partitions

    def run(self) -> None:
        """Runs on all objects of ``object_type`` in each partition between
        ``start_partition_id`` (inclusive) and ``end_partition_id`` (exclusive).
        """
        counter: Iterable[int] = count()
        if self.limit:
            counter = islice(counter, 0, self.limit)
        for _, partition_id in zip(
            counter, self.db.checked_partition_iter_next(self.config_id)
        ):
            logger.debug(
                "Processing %s partition %d/%d",
                self.object_type,
                partition_id,
                self.nb_partitions,
            )

            self._check_partition(self.object_type, partition_id)

            self.db.checked_partition_upsert(
                self.config_id,
                partition_id,
            )

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(psycopg2.OperationalError),
        wait=tenacity.wait_random_exponential(min=10, max=180),
    )
    def _check_partition(
        self, object_type: swhids.ObjectType, partition_id: int
    ) -> None:
        "Retryable method checking objects in partition."
        return self.check_partition(object_type, partition_id)

    @abstractmethod
    def check_partition(
        self, object_type: swhids.ObjectType, partition_id: int
    ) -> None:
        """Abstract method that derived classes must implement to check objects
        in partition."""
        pass



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from swh.scrubber.db import ScrubberDb


def get_scrubber_db(cls: str, **kwargs) -> ScrubberDb:
    from swh.core.config import get_swh_backend_module

    _, BackendCls = get_swh_backend_module("scrubber", cls)
    assert BackendCls is not None
    return BackendCls(**kwargs)


get_datastore = get_scrubber_db



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
import json

import attr
import pytest

from swh.journal.serializers import kafka_to_value
from swh.model.swhids import CoreSWHID, ObjectType
from swh.model.tests import swh_model_data
from swh.scrubber.objstorage_checker import (
    ObjectStorageCheckerFromJournal,
    ObjectStorageCheckerFromStoragePartition,
    get_objstorage_datastore,
)

from .storage_checker_tests import assert_checked_ranges

EXPECTED_PARTITIONS = {
    (ObjectType.CONTENT, 0, 4),
    (ObjectType.CONTENT, 1, 4),
    (ObjectType.CONTENT, 2, 4),
    (ObjectType.CONTENT, 3, 4),
}


@pytest.fixture
def datastore(swh_objstorage_config):
    return get_objstorage_datastore(swh_objstorage_config)


@pytest.fixture
def objstorage_checker_partition(swh_storage, swh_objstorage, scrubber_db, datastore):
    nb_partitions = len(EXPECTED_PARTITIONS)
    config_id = scrubber_db.config_add(
        "cfg_objstorage_checker_partition", datastore, ObjectType.CONTENT, nb_partitions
    )
    return ObjectStorageCheckerFromStoragePartition(
        scrubber_db, config_id, swh_storage, swh_objstorage
    )


@pytest.fixture
def objstorage_checker_journal(
    journal_client_config, swh_objstorage, scrubber_db, datastore
):
    config_id = scrubber_db.config_add(
        "cfg_objstorage_checker_journal", datastore, ObjectType.CONTENT, nb_partitions=1
    )
    return ObjectStorageCheckerFromJournal(
        scrubber_db, config_id, journal_client_config, swh_objstorage
    )


def test_objstorage_checker_partition_no_corruption(
    swh_storage, swh_objstorage, objstorage_checker_partition
):
    swh_storage.content_add(swh_model_data.CONTENTS)
    swh_objstorage.add_batch((c.hashes(), c.data) for c in swh_model_data.CONTENTS)

    objstorage_checker_partition.run()

    scrubber_db = objstorage_checker_partition.db
    assert list(scrubber_db.corrupt_object_iter()) == []

    assert_checked_ranges(
        scrubber_db,
        [(ObjectType.CONTENT, objstorage_checker_partition.config_id)],
        EXPECTED_PARTITIONS,
    )


@pytest.mark.parametrize("missing_idx", range(0, len(swh_model_data.CONTENTS), 5))
def test_objstorage_checker_partition_missing_content(
    swh_storage, swh_objstorage, objstorage_checker_partition, missing_idx
):
    contents = list(swh_model_data.CONTENTS)
    swh_storage.content_add(contents)
    swh_objstorage.add_batch(
        (c.hashes(), c.data) for i, c in enumerate(contents) if i != missing_idx
    )

    before_date = datetime.now(tz=timezone.utc)
    objstorage_checker_partition.run()
    after_date = datetime.now(tz=timezone.utc)

    scrubber_db = objstorage_checker_partition.db

    missing_objects = list(scrubber_db.missing_object_iter())
    assert len(missing_objects) == 1
    assert missing_objects[0].id == contents[missing_idx].swhid()
    assert missing_objects[0].config.datastore == objstorage_checker_partition.datastore
    assert (
        before_date - timedelta(seconds=5)
        <= missing_objects[0].first_occurrence
        <= after_date + timedelta(seconds=5)
    )

    assert_checked_ranges(
        scrubber_db,
        [(ObjectType.CONTENT, objstorage_checker_partition.config_id)],
        EXPECTED_PARTITIONS,
        before_date,
        after_date,
    )


@pytest.mark.parametrize("corrupt_idx", range(0, len(swh_model_data.CONTENTS), 5))
def test_objstorage_checker_partition_corrupt_content(
    swh_storage, swh_objstorage, objstorage_checker_partition, corrupt_idx
):
    contents = list(swh_model_data.CONTENTS)
    contents[corrupt_idx] = attr.evolve(contents[corrupt_idx], sha1_git=b"\x00" * 20)
    swh_storage.content_add(contents)
    swh_objstorage.add_batch((c.hashes(), c.data) for c in contents)

    before_date = datetime.now(tz=timezone.utc)
    objstorage_checker_partition.run()
    after_date = datetime.now(tz=timezone.utc)

    scrubber_db = objstorage_checker_partition.db

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == CoreSWHID.from_string(
        "swh:1:cnt:0000000000000000000000000000000000000000"
    )
    assert corrupt_objects[0].config.datastore == objstorage_checker_partition.datastore
    assert (
        before_date - timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + timedelta(seconds=5)
    )

    corrupted_content = contents[corrupt_idx].to_dict()
    corrupted_content.pop("data")
    assert kafka_to_value(corrupt_objects[0].object_) == corrupted_content

    assert_checked_ranges(
        scrubber_db,
        [(ObjectType.CONTENT, objstorage_checker_partition.config_id)],
        EXPECTED_PARTITIONS,
        before_date,
        after_date,
    )


def test_objstorage_checker_journal_contents_no_corruption(
    scrubber_db,
    journal_writer,
    journal_client_config,
    objstorage_checker_journal,
):
    journal_writer.write_additions("content", swh_model_data.CONTENTS)

    gid = journal_client_config["group_id"] + "_"

    object_type = "content"
    journal_client_config["group_id"] = gid + object_type

    objstorage_checker_journal.objstorage.add_batch(
        (c.hashes(), c.data) for c in swh_model_data.CONTENTS
    )
    objstorage_checker_journal.run()
    objstorage_checker_journal.journal_client.close()

    assert list(scrubber_db.corrupt_object_iter()) == []


@pytest.mark.parametrize("corrupt_idx", range(0, len(swh_model_data.CONTENTS), 5))
def test_objstorage_checker_journal_corrupt_content(
    scrubber_db,
    journal_writer,
    objstorage_checker_journal,
    swh_objstorage_config,
    corrupt_idx,
):
    contents = list(swh_model_data.CONTENTS)
    contents[corrupt_idx] = attr.evolve(contents[corrupt_idx], sha1_git=b"\x00" * 20)

    journal_writer.write_additions("content", contents)

    before_date = datetime.now(tz=timezone.utc)

    objstorage_checker_journal.objstorage.add_batch(
        (c.hashes(), c.data) for c in contents
    )
    objstorage_checker_journal.run()
    after_date = datetime.now(tz=timezone.utc)

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == CoreSWHID.from_string(
        "swh:1:cnt:0000000000000000000000000000000000000000"
    )
    assert corrupt_objects[0].config.datastore.package == "objstorage"
    assert corrupt_objects[0].config.datastore.cls == swh_objstorage_config.pop("cls")
    assert corrupt_objects[0].config.datastore.instance == json.dumps(
        swh_objstorage_config
    )
    assert (
        before_date - timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + timedelta(seconds=5)
    )
    corrupted_content = contents[corrupt_idx].to_dict()
    corrupted_content.pop("data")
    assert kafka_to_value(corrupt_objects[0].object_) == corrupted_content


@pytest.mark.parametrize("missing_idx", range(0, len(swh_model_data.CONTENTS), 5))
def test_objstorage_checker_journal_missing_content(
    scrubber_db,
    journal_writer,
    objstorage_checker_journal,
    missing_idx,
):
    contents = list(swh_model_data.CONTENTS)

    journal_writer.write_additions("content", contents)

    before_date = datetime.now(tz=timezone.utc)

    objstorage_checker_journal.objstorage.add_batch(
        (c.hashes(), c.data) for i, c in enumerate(contents) if i != missing_idx
    )
    objstorage_checker_journal.run()
    after_date = datetime.now(tz=timezone.utc)

    missing_objects = list(scrubber_db.missing_object_iter())
    assert len(missing_objects) == 1
    assert missing_objects[0].id == contents[missing_idx].swhid()
    assert missing_objects[0].config.datastore == objstorage_checker_journal.datastore
    assert (
        before_date - timedelta(seconds=5)
        <= missing_objects[0].first_occurrence
        <= after_date + timedelta(seconds=5)
    )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from .objstorage_checker_tests import *  # noqa

# Use cassandra storage and an objstorage with memory backend to run
# the tests


@pytest.fixture
def swh_storage_backend_config(swh_storage_cassandra_backend_config):
    return swh_storage_cassandra_backend_config


@pytest.fixture
def swh_objstorage_config():
    return {"cls": "memory"}



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import tempfile
import traceback
from unittest.mock import MagicMock, call

from click.testing import CliRunner
import yaml

from swh.model.swhids import CoreSWHID
from swh.scrubber.cli import scrubber_cli_group
from swh.scrubber.storage_checker import postgresql_storage_db


def assert_result(result):
    if result.exception:
        assert result.exit_code == 0, (
            "Unexpected exception: "
            f"{''.join(traceback.format_tb(result.exc_info[2]))}"
            f"\noutput: {result.output}"
        )
    else:
        assert result.exit_code == 0, f"Unexpected output: {result.output}"


def invoke(
    scrubber_db,
    args,
    storage=None,
    objstorage=None,
    kafka_server=None,
    kafka_prefix=None,
    kafka_consumer_group=None,
):
    runner = CliRunner()

    config = {
        "scrubber": {"cls": "postgresql", "db": scrubber_db.conn.dsn},
        "graph": {"url": "http://graph.example.org:5009/"},
    }
    if storage:
        with postgresql_storage_db(storage) as db:
            config["storage"] = {
                "cls": "postgresql",
                "db": db.conn.dsn,
                "objstorage": {"cls": "memory"},
            }
    if objstorage:
        config["objstorage"] = {"cls": "memory"}

    assert (
        (kafka_server is None)
        == (kafka_prefix is None)
        == (kafka_consumer_group is None)
    )
    if kafka_server:
        config["journal"] = dict(
            cls="kafka",
            brokers=kafka_server,
            group_id=kafka_consumer_group,
            prefix=kafka_prefix,
            on_eof="stop",
        )

    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        yaml.dump(config, config_fd)
        config_fd.seek(0)
        args = ["-C" + config_fd.name] + list(args)
        result = runner.invoke(scrubber_cli_group, args, catch_exceptions=False)
    return result


def test_help_main(mocker, scrubber_db, swh_storage):
    result = invoke(
        scrubber_db,
        [
            "--help",
        ],
    )
    assert_result(result)
    output = result.output.splitlines(keepends=False)
    msg = "Usage: scrubber [OPTIONS] COMMAND [ARGS]..."
    assert output[0] == msg
    assert "Commands:" in output
    commands = [cmd.split()[0] for cmd in output[output.index("Commands:") + 1 :]]
    assert commands == ["check", "fix", "locate"]


def test_help_check(mocker, scrubber_db, swh_storage):
    result = invoke(
        scrubber_db,
        [
            "check",
            "--help",
        ],
    )
    assert_result(result)
    output = result.output.splitlines(keepends=False)
    msg = "Usage: scrubber check [OPTIONS] COMMAND [ARGS]..."
    assert output[0] == msg
    assert "Commands:" in output
    commands = [cmd.split()[0] for cmd in output[output.index("Commands:") + 1 :]]
    # With older click version (e.g. 7.0-1), the text wrapping can be different,
    # resulting in some docstring text included in this command list, so check we find
    # the expected commands instead
    for command in ["init", "list", "run", "stalled"]:
        assert command in commands
    for command in ["storage", "journal"]:
        assert command not in commands

    # without a config file, --help should still work
    result = CliRunner().invoke(
        scrubber_cli_group, ["check", "--help"], catch_exceptions=False
    )
    output = result.output.splitlines(keepends=False)
    msg = "Usage: scrubber check [OPTIONS] COMMAND [ARGS]..."
    assert output[0] == msg
    assert "Commands:" in output
    commands = [cmd.split()[0] for cmd in output[output.index("Commands:") + 1 :]]
    # With older click version (e.g. 7.0-1), the text wrapping can be different,
    # resulting in some docstring text included in this command list, so check we find
    # the expected commands instead
    for command in ["init", "list", "run", "stalled"]:
        assert command in commands


def test_check_init_storage(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)
    msg = "Created configuration cfg1 [1] for checking snapshot in postgresql storage"
    assert result.output.strip() == msg

    # error: cfg name already exists
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "8",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 1, result.output
    msg = "Error: Configuration cfg1 already exists"
    assert result.output.strip() == msg


def test_check_init_storage_flags(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    arg_list = [
        "check",
        "init",
        "storage",
        "--object-type",
        "snapshot",
        "--nb-partitions",
        "4",
        "--name",
    ]

    name = "cfg1"
    result = invoke(
        scrubber_db,
        arg_list + [name],
        storage=swh_storage,
    )
    assert_result(result)

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is True

    name = "cfg2"
    result = invoke(
        scrubber_db,
        arg_list + [name, "--no-check-references"],
        storage=swh_storage,
    )
    assert_result(result)

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is False

    name = "cfg3"
    result = invoke(
        scrubber_db,
        arg_list + [name, "--no-check-hashes"],
        storage=swh_storage,
    )
    assert_result(result)

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is False
    assert cfg_entry.check_references is True


def test_check_init_objstorage(mocker, scrubber_db, swh_storage, swh_objstorage):
    config_name = "cfg1"
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "objstorage",
            "--object-type",
            "content",
            "--nb-partitions",
            "4",
            "--name",
            config_name,
        ],
        storage=swh_storage,
        objstorage=swh_objstorage,
    )
    assert_result(result)
    msg = f"Created configuration {config_name} [1] for checking content in memory objstorage"
    assert result.output.strip() == msg

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(config_name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is True

    # error: config name already exists
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "objstorage",
            "--object-type",
            "content",
            "--nb-partitions",
            "8",
            "--name",
            config_name,
        ],
        storage=swh_storage,
        objstorage=swh_objstorage,
    )
    assert result.exit_code == 1, result.output
    msg = f"Error: Configuration {config_name} already exists"
    assert result.output.strip() == msg

    # error: missing objstorage config
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "objstorage",
            "--object-type",
            "content",
            "--nb-partitions",
            "8",
            "--name",
            config_name,
        ],
        storage=swh_storage,
    )
    assert result.exit_code == 1, result.output
    msg = "Error: You must have an object storage configured in your config file."
    assert result.output.strip() == msg


def test_check_init_journal_flags(
    mocker, scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group
):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    arg_list = [
        "check",
        "init",
        "journal",
        "--object-type",
        "snapshot",
        "--name",
    ]

    name = "cfg1"
    result = invoke(
        scrubber_db,
        arg_list + [name],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert_result(result)

    cfg_entry = scrubber_db.config_get(scrubber_db.config_get_by_name(name))
    assert cfg_entry.check_hashes is True
    assert cfg_entry.check_references is False


def test_check_run_ko(mocker, scrubber_db, swh_storage):
    # using the config id instead of the config name
    result = invoke(scrubber_db, ["check", "run"], storage=swh_storage)
    assert result.exit_code == 2, result.output
    assert "Error: A valid configuration name/id must be given." in result.output


def test_check_run_storage(mocker, scrubber_db, swh_storage):
    storage_checker = MagicMock()
    StorageChecker = mocker.patch(
        "swh.scrubber.storage_checker.StorageChecker", return_value=storage_checker
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)
    msg = "Created configuration cfg1 [1] for checking snapshot in postgresql storage"
    assert result.output.strip() == msg

    result = invoke(scrubber_db, ["check", "run", "cfg1"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""

    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)
    StorageChecker.assert_called_once_with(
        db=scrubber_db,
        config_id=1,
        storage=StorageChecker.mock_calls[0][2]["storage"],
        limit=0,
    )
    assert storage_checker.method_calls == [call.run()]

    # using the config id instead of the config name
    result = invoke(
        scrubber_db, ["check", "run", "--config-id", "1"], storage=swh_storage
    )
    assert_result(result)
    assert result.output == ""


def test_check_run_objstorage_partition(
    mocker, scrubber_db, swh_storage, swh_objstorage
):
    config_name = "cfg1"
    objstorage_checker = MagicMock()
    ObjectStorageCheckerFromStoragePartition = mocker.patch(
        "swh.scrubber.objstorage_checker.ObjectStorageCheckerFromStoragePartition",
        return_value=objstorage_checker,
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "objstorage",
            "--object-type",
            "content",
            "--nb-partitions",
            "4",
            "--name",
            config_name,
        ],
        storage=swh_storage,
        objstorage=swh_objstorage,
    )
    assert_result(result)
    msg = f"Created configuration {config_name} [1] for checking content in memory objstorage"
    assert result.output.strip() == msg

    result = invoke(
        scrubber_db,
        ["check", "run", config_name],
        storage=swh_storage,
        objstorage=swh_objstorage,
    )
    assert_result(result)
    assert result.output == ""

    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)
    ObjectStorageCheckerFromStoragePartition.assert_called_once_with(
        db=scrubber_db,
        config_id=1,
        storage=ObjectStorageCheckerFromStoragePartition.mock_calls[0][2]["storage"],
        objstorage=ObjectStorageCheckerFromStoragePartition.mock_calls[0][2][
            "objstorage"
        ],
        limit=0,
    )
    assert objstorage_checker.method_calls == [call.run()]

    # using the config id instead of the config name
    result = invoke(
        scrubber_db,
        ["check", "run", "--config-id", "1"],
        storage=swh_storage,
        objstorage=swh_objstorage,
    )
    assert_result(result)
    assert result.output == ""


def test_check_run_objstorage_journal(
    mocker,
    scrubber_db,
    kafka_server,
    kafka_prefix,
    kafka_consumer_group,
    swh_objstorage,
):
    config_name = "cfg1"
    journal_checker = MagicMock()
    ObjectStorageCheckerFromJournal = mocker.patch(
        "swh.scrubber.objstorage_checker.ObjectStorageCheckerFromJournal",
        return_value=journal_checker,
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "objstorage",
            "--object-type",
            "content",
            "--name",
            config_name,
        ],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
        objstorage=swh_objstorage,
    )
    assert_result(result)
    msg = f"Created configuration {config_name} [1] for checking content in memory objstorage"
    assert result.output.strip() == msg

    result = invoke(
        scrubber_db,
        ["check", "run", "--use-journal", config_name],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
        objstorage=swh_objstorage,
    )
    assert_result(result)
    assert result.output == ""

    assert get_scrubber_db.call_count == 2
    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)

    ObjectStorageCheckerFromJournal.assert_called_once_with(
        db=scrubber_db,
        journal_client_config={
            "brokers": kafka_server,
            "cls": "kafka",
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
            "on_eof": "stop",
        },
        config_id=1,
        objstorage=ObjectStorageCheckerFromJournal.mock_calls[0][2]["objstorage"],
    )
    assert journal_checker.method_calls == [call.run()]


def test_check_run_journal(
    mocker, scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group
):
    journal_checker = MagicMock()
    JournalChecker = mocker.patch(
        "swh.scrubber.journal_checker.JournalChecker", return_value=journal_checker
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "journal",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert_result(result)
    msg = "Created configuration cfg1 [1] for checking snapshot in kafka journal"
    assert result.output.strip() == msg

    result = invoke(
        scrubber_db,
        ["check", "run", "cfg1"],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert_result(result)
    assert result.output == ""

    assert get_scrubber_db.call_count == 2
    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)

    JournalChecker.assert_called_once_with(
        db=scrubber_db,
        journal_client_config={
            "brokers": kafka_server,
            "cls": "kafka",
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
            "on_eof": "stop",
        },
        config_id=1,
    )
    assert journal_checker.method_calls == [call.run()]


def test_check_list(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""
    with swh_storage.db() as db:
        dsn = db.conn.dsn

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)

    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert_result(result)
    expected = f"[1] cfg1: snapshot, 4, storage:postgresql ({dsn})\n"
    assert result.output == expected, result.output


def test_check_stalled(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)

    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a partition started 20mn ago
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition VALUES (1, 0, now() - '20m'::interval, NULL);"
        )

    # there are no existing completed partition, defaults to 1h to be considered as stalled
    # so a partition just added is not stalled
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a partition started 2 hours from now
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition VALUES (1, 1, now() - '2h'::interval, NULL);"
        )
    # it is considered as stalled by default
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
1:	stuck since today (2 hours)
"""
    assert result.output == expected, result.output

    # explicitly specify a delay > 2h to be considered as stelles: no one stalled
    result = invoke(
        scrubber_db, ["check", "stalled", "--for", "8000", "cfg1"], storage=swh_storage
    )
    assert_result(result)
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a transaction that took 5mn to run
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 2, now() - '2h'::interval, now() - '1h55m'::interval);"
        )
    # so now both partitions 0 and 1 should be considered a stalled
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
0:	stuck since today (20 minutes)
1:	stuck since today (2 hours)
"""
    assert result.output == expected, result.output


def test_check_running(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)

    result = invoke(scrubber_db, ["check", "running", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = "No running partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a partition started 20mn ago
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition VALUES (1, 0, now() - '20m'::interval, NULL);"
        )
    result = invoke(scrubber_db, ["check", "running", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = """\
Running partitions for cfg1 [id=1, type=snapshot]:
0:	running since today (20 minutes)
"""
    assert result.output == expected, result.output

    # insert another partition started 2 hours ago
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition VALUES (1, 1, now() - '2h'::interval, NULL);"
        )
    result = invoke(scrubber_db, ["check", "running", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = """\
Running partitions for cfg1 [id=1, type=snapshot]:
0:	running since today (20 minutes)
1:	running since today (2 hours)
"""
    assert result.output == expected, result.output

    # insert another partition whose scrubbing did not start yet
    with scrubber_db.transaction() as cur:
        cur.execute("INSERT INTO checked_partition VALUES (1, 2, NULL, NULL);")

    result = invoke(scrubber_db, ["check", "running", "cfg1"], storage=swh_storage)

    # not scrubbed partition should not be displayed
    assert_result(result)
    assert result.output == expected, result.output


def test_check_stats(mocker, scrubber_db, swh_storage):
    from swh.scrubber.storage_checker import get_datastore

    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""

    for otype in ("snapshot", "revision", "release"):
        result = invoke(
            scrubber_db,
            [
                "check",
                "init",
                "storage",
                "--object-type",
                otype,
                "--nb-partitions",
                "4",
                "--name",
                f"cfg_{otype}",
            ],
            storage=swh_storage,
        )
        assert_result(result)

    result = invoke(
        scrubber_db, ["check", "stats", "cfg_snapshot"], storage=swh_storage
    )
    assert_result(result)

    for otype in ("snapshot", "revision", "release"):
        result = invoke(
            scrubber_db,
            ["check", "stats", "--json", f"cfg_{otype}"],
            storage=swh_storage,
        )
        stats = json.loads(result.output)
        assert stats == {
            "config": {
                "name": f"cfg_{otype}",
                "datastore": {
                    "package": "storage",
                    "cls": "postgresql",
                    "instance": get_datastore(swh_storage).instance,
                },
                "object_type": otype,
                "nb_partitions": 4,
                "check_hashes": True,
                "check_references": True,
            },
            "min_duration": 0,
            "max_duration": 0,
            "avg_duration": 0,
            "checked_partition": 0,
            "running_partition": 0,
            "missing_object": 0,
            "missing_object_reference": 0,
            "corrupt_object": 0,
        }

    with scrubber_db.transaction() as cur:
        # insert a pair of checked partitions
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 0, now() - '40m'::interval, now() - '20m'::interval)"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 1, now() - '20m'::interval, now() - '10m'::interval)"
        )
        # and a pair of running ones
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 2, now() - '10m'::interval, NULL)"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 3, now() - '20m'::interval, NULL)"
        )
        # also add a checked partitions for another config entry
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (2, 0, now() - '40m'::interval, now() - '20m'::interval)"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (2, 1, now() - '40m'::interval, now() - '20m'::interval)"
        )
        # and add a running checker for another config entry
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (2, 3, now() - '20m'::interval, NULL)"
        )
    result = invoke(
        scrubber_db, ["check", "stats", "-j", "cfg_snapshot"], storage=swh_storage
    )
    assert_result(result)
    stats = json.loads(result.output)
    assert stats["config"]["name"] == "cfg_snapshot"
    assert stats["min_duration"] == 600
    assert stats["max_duration"] == 1200
    assert stats["avg_duration"] == 900
    assert stats["checked_partition"] == 2
    assert stats["running_partition"] == 2


def test_check_reset(mocker, scrubber_db, swh_storage):
    mocker.patch("swh.scrubber.get_scrubber_db", return_value=scrubber_db)
    result = invoke(scrubber_db, ["check", "list"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""

    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)

    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = "No stuck partition found for cfg1 [id=1, type=snapshot]\n"
    assert result.output == expected, result.output

    # insert a few partitions
    with scrubber_db.transaction() as cur:
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 0, now() - '20m'::interval, NULL);"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 1, now() - '2h'::interval, NULL);"
        )
        cur.execute(
            "INSERT INTO checked_partition "
            "VALUES (1, 2, now() - '2h'::interval, now() - '1h55m'::interval);"
        )

    # partitions 0 and 1 are considered as stalled
    result = invoke(scrubber_db, ["check", "stalled", "cfg1"], storage=swh_storage)
    assert_result(result)
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
0:	stuck since today (20 minutes)
1:	stuck since today (2 hours)
"""
    assert result.output == expected, result.output

    # let's reset them
    result = invoke(
        scrubber_db, ["check", "stalled", "--reset", "cfg1"], storage=swh_storage
    )
    assert_result(result)
    expected = """\
Stuck partitions for cfg1 [id=1, type=snapshot]:
0:	stuck since today (20 minutes)
	partition reset
1:	stuck since today (2 hours)
	partition reset
"""  # noqa: W191,E101
    assert result.output == expected, result.output
    with scrubber_db.transaction() as cur:
        cur.execute(
            "SELECT partition_id, end_date "
            "FROM checked_partition "
            "WHERE config_id=1 AND start_date is NULL"
        )
        assert cur.fetchall() == [(0, None), (1, None)]

    # for good measure, check the next few selected partitions, expected 0, 1 and 3
    assert next(scrubber_db.checked_partition_iter_next(1)) == 0
    assert next(scrubber_db.checked_partition_iter_next(1)) == 1
    assert next(scrubber_db.checked_partition_iter_next(1)) == 3


def test_locate_origins(mocker, scrubber_db, swh_storage, naive_graph_client):
    origin_locator = MagicMock()
    OriginLocator = mocker.patch(
        "swh.scrubber.origin_locator.OriginLocator", return_value=origin_locator
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    mocker.patch(
        "swh.graph.http_client.RemoteGraphClient",
        return_value=naive_graph_client,
    )

    result = invoke(scrubber_db, ["locate"], storage=swh_storage)
    assert_result(result)
    assert result.output == ""

    get_scrubber_db.assert_called_once_with(cls="postgresql", db=scrubber_db.conn.dsn)
    OriginLocator.assert_called_once_with(
        db=scrubber_db,
        storage=OriginLocator.mock_calls[0][2]["storage"],
        graph=OriginLocator.mock_calls[0][2]["graph"],
        start_object=CoreSWHID.from_string("swh:1:cnt:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:snp:" + "ff" * 20),
    )
    assert origin_locator.method_calls == [call.run()]


def test_fix_objects(mocker, scrubber_db):
    fixer = MagicMock()
    Fixer = mocker.patch("swh.scrubber.fixer.Fixer", return_value=fixer)
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(scrubber_db, ["fix"])
    assert_result(result)
    assert result.output == ""

    get_scrubber_db.assert_called_once_with(cls="postgresql", db=scrubber_db.conn.dsn)
    Fixer.assert_called_once_with(
        db=scrubber_db,
        start_object=CoreSWHID.from_string("swh:1:cnt:" + "00" * 20),
        end_object=CoreSWHID.from_string("swh:1:snp:" + "ff" * 20),
    )
    assert fixer.method_calls == [call.run()]


# deprecated commands
def test_check_storage(mocker, scrubber_db, swh_storage):
    storage_checker = MagicMock()
    StorageChecker = mocker.patch(
        "swh.scrubber.storage_checker.StorageChecker", return_value=storage_checker
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "storage",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        storage=swh_storage,
    )
    assert_result(result)
    msg = "Created configuration cfg1 [1] for checking snapshot in postgresql storage"
    assert result.output.strip() == msg

    result = invoke(scrubber_db, ["check", "storage", "cfg1"], storage=swh_storage)
    assert_result(result)
    assert (
        result.output.strip()
        == "DeprecationWarning: The command 'storage' is deprecated."
    )

    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)
    StorageChecker.assert_called_once_with(
        db=scrubber_db,
        config_id=1,
        storage=StorageChecker.mock_calls[0][2]["storage"],
        limit=0,
    )
    assert storage_checker.method_calls == [call.run()]

    # using the config id instead of the config name
    result = invoke(
        scrubber_db, ["check", "storage", "--config-id", "1"], storage=swh_storage
    )
    assert_result(result)
    assert (
        result.output.strip()
        == "DeprecationWarning: The command 'storage' is deprecated."
    )


def test_check_journal(
    mocker, scrubber_db, kafka_server, kafka_prefix, kafka_consumer_group
):
    journal_checker = MagicMock()
    JournalChecker = mocker.patch(
        "swh.scrubber.journal_checker.JournalChecker", return_value=journal_checker
    )
    get_scrubber_db = mocker.patch(
        "swh.scrubber.get_scrubber_db", return_value=scrubber_db
    )
    result = invoke(
        scrubber_db,
        [
            "check",
            "init",
            "journal",
            "--object-type",
            "snapshot",
            "--nb-partitions",
            "4",
            "--name",
            "cfg1",
        ],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert_result(result)
    msg = "Created configuration cfg1 [1] for checking snapshot in kafka journal"
    assert result.output.strip() == msg

    result = invoke(
        scrubber_db,
        ["check", "journal", "cfg1"],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
    )
    assert_result(result)
    assert (
        result.output.strip()
        == "DeprecationWarning: The command 'journal' is deprecated."
    )

    assert get_scrubber_db.call_count == 2
    get_scrubber_db.assert_called_with(cls="postgresql", db=scrubber_db.conn.dsn)

    JournalChecker.assert_called_once_with(
        db=scrubber_db,
        journal_client_config={
            "brokers": kafka_server,
            "cls": "kafka",
            "group_id": kafka_consumer_group,
            "prefix": kafka_prefix,
            "on_eof": "stop",
        },
        config_id=1,
    )
    assert journal_checker.method_calls == [call.run()]



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import hashlib

import attr
import pytest

from swh.journal.serializers import kafka_to_value
from swh.model import model, swhids
from swh.model.swhids import ObjectType
from swh.model.tests import swh_model_data
from swh.scrubber.db import Datastore
from swh.scrubber.journal_checker import JournalChecker, get_datastore


@pytest.fixture
def datastore(journal_client_config) -> Datastore:
    return get_datastore(journal_client_config)


def test_no_corruption(
    scrubber_db,
    datastore,
    journal_writer,
    journal_client_config,
):
    journal_writer.write_additions("directory", swh_model_data.DIRECTORIES)
    journal_writer.write_additions("revision", swh_model_data.REVISIONS)
    journal_writer.write_additions("release", swh_model_data.RELEASES)
    journal_writer.write_additions("snapshot", swh_model_data.SNAPSHOTS)

    gid = journal_client_config["group_id"] + "_"

    for object_type in ("directory", "revision", "release", "snapshot"):
        journal_client_config["group_id"] = gid + object_type
        config_id = scrubber_db.config_add(
            name=f"cfg_{object_type}",
            datastore=datastore,
            object_type=getattr(ObjectType, object_type.upper()),
            nb_partitions=1,
            check_references=False,
        )
        jc = JournalChecker(
            db=scrubber_db,
            config_id=config_id,
            journal_client_config=journal_client_config,
        )
        jc.run()
        jc.journal_client.close()

    assert list(scrubber_db.corrupt_object_iter()) == []


@pytest.mark.parametrize("corrupt_idx", range(len(swh_model_data.SNAPSHOTS)))
def test_corrupt_snapshot(
    scrubber_db,
    datastore,
    journal_writer,
    journal_client_config,
    corrupt_idx,
):
    config_id = scrubber_db.config_add(
        name="cfg_snapshot",
        datastore=datastore,
        object_type=ObjectType.SNAPSHOT,
        nb_partitions=1,
        check_references=False,
    )
    snapshots = list(swh_model_data.SNAPSHOTS)
    snapshots[corrupt_idx] = attr.evolve(snapshots[corrupt_idx], id=b"\x00" * 20)

    journal_writer.write_additions("snapshot", snapshots)

    before_date = datetime.datetime.now(tz=datetime.timezone.utc)
    JournalChecker(
        db=scrubber_db,
        config_id=config_id,
        journal_client_config=journal_client_config,
    ).run()
    after_date = datetime.datetime.now(tz=datetime.timezone.utc)

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 1
    assert corrupt_objects[0].id == swhids.CoreSWHID.from_string(
        "swh:1:snp:0000000000000000000000000000000000000000"
    )
    assert corrupt_objects[0].config.datastore.package == "journal"
    assert corrupt_objects[0].config.datastore.cls == "kafka"
    assert (
        before_date - datetime.timedelta(seconds=5)
        <= corrupt_objects[0].first_occurrence
        <= after_date + datetime.timedelta(seconds=5)
    )
    assert (
        kafka_to_value(corrupt_objects[0].object_) == snapshots[corrupt_idx].to_dict()
    )


def test_corrupt_snapshots(
    scrubber_db,
    datastore,
    journal_writer,
    journal_client_config,
):
    config_id = scrubber_db.config_add(
        name="cfg_snapshot",
        datastore=datastore,
        object_type=ObjectType.SNAPSHOT,
        nb_partitions=1,
        check_references=False,
    )
    snapshots = list(swh_model_data.SNAPSHOTS)
    for i in (0, 1):
        snapshots[i] = attr.evolve(snapshots[i], id=bytes([i]) * 20)

    journal_writer.write_additions("snapshot", snapshots)

    JournalChecker(
        db=scrubber_db,
        config_id=config_id,
        journal_client_config=journal_client_config,
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 2
    assert {co.id for co in corrupt_objects} == {
        swhids.CoreSWHID.from_string(swhid)
        for swhid in [
            "swh:1:snp:0000000000000000000000000000000000000000",
            "swh:1:snp:0101010101010101010101010101010101010101",
        ]
    }


def test_duplicate_directory_entries(
    scrubber_db,
    datastore,
    journal_writer,
    kafka_prefix,
    journal_client_config,
):
    config_id = scrubber_db.config_add(
        name="cfg_directory",
        datastore=datastore,
        object_type=ObjectType.DIRECTORY,
        nb_partitions=1,
        check_references=False,
    )
    directory = model.Directory(
        entries=(
            model.DirectoryEntry(
                name=b"filename", type="file", target=b"\x01" * 20, perms=0
            ),
        )
    )

    # has duplicate entries and wrong hash
    corrupt_directory = {
        "id": b"\x00" * 20,
        "entries": [
            {"name": b"filename", "type": "file", "target": b"\x01" * 20, "perms": 0},
            {"name": b"filename", "type": "file", "target": b"\x02" * 20, "perms": 0},
        ],
    }

    # has duplicate entries but correct hash
    raw_manifest = (
        b"tree 62\x00"
        + b"0 filename\x00"
        + b"\x01" * 20
        + b"0 filename\x00"
        + b"\x02" * 20
    )
    dup_directory = {
        "id": hashlib.sha1(raw_manifest).digest(),
        "entries": corrupt_directory["entries"],
        "raw_manifest": raw_manifest,
    }

    journal_writer.send(f"{kafka_prefix}.directory", directory.id, directory.to_dict())
    journal_writer.send(
        f"{kafka_prefix}.directory", corrupt_directory["id"], corrupt_directory
    )
    journal_writer.send(f"{kafka_prefix}.directory", dup_directory["id"], dup_directory)

    JournalChecker(
        db=scrubber_db,
        config_id=config_id,
        journal_client_config=journal_client_config,
    ).run()

    corrupt_objects = list(scrubber_db.corrupt_object_iter())
    assert len(corrupt_objects) == 2
    assert {co.id for co in corrupt_objects} == {
        swhids.CoreSWHID.from_string(swhid)
        for swhid in [
            "swh:1:dir:0000000000000000000000000000000000000000",
            f"swh:1:dir:{dup_directory['id'].hex()}",
        ]
    }


def test_check_references_raises(
    scrubber_db,
    datastore,
    journal_client_config,
):
    config_id = scrubber_db.config_add(
        name="cfg_snapshot",
        datastore=datastore,
        object_type=ObjectType.SNAPSHOT,
        nb_partitions=1,
        check_references=True,
    )

    with pytest.raises(ValueError):
        JournalChecker(
            db=scrubber_db,
            config_id=config_id,
            journal_client_config=journal_client_config,
        )



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.scrubber import get_scrubber_db
from swh.scrubber.db import ScrubberDb


def test_get_scrubber_db(scrubber_db):
    assert isinstance(scrubber_db, ScrubberDb)


@pytest.mark.parametrize("clz", ["something", "anything"])
def test_get_scrubber_db_raise(clz):
    with pytest.raises(ValueError, match="Unknown"):
        get_scrubber_db(clz, db="service=scrubber-db")



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from unittest.mock import patch

from psycopg2.errors import UniqueViolation
import pytest

from swh.model.swhids import ObjectType
from swh.scrubber.db import Datastore, ScrubberDb

from .conftest import NB_PARTITIONS, OBJECT_TYPE

DATE = datetime.datetime(2022, 10, 4, 12, 1, 23, tzinfo=datetime.timezone.utc)
ONE_MINUTE = datetime.timedelta(minutes=1)
ONE_DAY = datetime.timedelta(days=1)


def test_config_add(datastore: Datastore, scrubber_db: ScrubberDb, config_id: int):
    cfg_snp = scrubber_db.config_add("cfg snp", datastore, ObjectType.SNAPSHOT, 42)
    assert cfg_snp == 2

    cfg_snp2 = scrubber_db.config_add("cfg snp 2", datastore, ObjectType.SNAPSHOT, 43)
    assert cfg_snp2 == 3

    # if not given, a config name is computed
    cfg_snp3 = scrubber_db.config_add(None, datastore, ObjectType.SNAPSHOT, 44)
    assert cfg_snp3 == 4
    assert (
        scrubber_db.config_get(cfg_snp3).name == "check_snapshot_44_storage_postgresql"
    )

    # XXX this is debatable: is there a good reason to allow 2 configs for the
    # same datastore and object type but different partition number, but not
    # for the same number of partitions?
    with pytest.raises(UniqueViolation):
        scrubber_db.config_add("cfg4", datastore, OBJECT_TYPE, NB_PARTITIONS)


def test_config_add_flags(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    id_cfg2 = scrubber_db.config_add("cfg snp", datastore, ObjectType.SNAPSHOT, 42)
    assert id_cfg2 == 2
    id_cfg3 = scrubber_db.config_add(
        "cfg3", datastore, ObjectType.SNAPSHOT, 43, check_hashes=False
    )
    assert id_cfg3 == 3
    id_cfg4 = scrubber_db.config_add(
        "cfg4", datastore, ObjectType.SNAPSHOT, 43, check_references=False
    )
    assert id_cfg4 == 4

    # but cannot add another config entry with the same name, ds, objtype and part
    # number but different flags
    with pytest.raises(UniqueViolation):
        scrubber_db.config_add(
            "cfg4", datastore, ObjectType.SNAPSHOT, 43, check_hashes=False
        )

    with pytest.raises(ValueError):
        scrubber_db.config_add(
            "cfg4",
            datastore,
            OBJECT_TYPE,
            NB_PARTITIONS,
            check_hashes=False,
            check_references=False,
        )


def test_config_get(datastore: Datastore, scrubber_db: ScrubberDb, config_id: int):
    id_cfg2 = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
    id_cfg3 = scrubber_db.config_add(
        "cfg3", datastore, ObjectType.SNAPSHOT, 43, False, True
    )
    id_cfg4 = scrubber_db.config_add(
        "cfg4", datastore, ObjectType.SNAPSHOT, 43, True, False
    )

    cfg2 = scrubber_db.config_get(id_cfg2)
    assert cfg2
    assert cfg2.check_hashes is True
    assert cfg2.check_references is True
    cfg3 = scrubber_db.config_get(id_cfg3)
    assert cfg3
    assert cfg3.check_hashes is False
    assert cfg3.check_references is True
    cfg4 = scrubber_db.config_get(id_cfg4)
    assert cfg4
    assert cfg4.check_hashes is True
    assert cfg4.check_references is False

    with pytest.raises(ValueError):
        scrubber_db.config_get(id_cfg4 + 1)


@pytest.fixture
def datastore2():
    return Datastore(package="storage", cls="postgresql", instance="service=swh-test-2")


@pytest.fixture
def config_id2(scrubber_db, datastore2):
    return scrubber_db.config_add(
        f"cfg_{OBJECT_TYPE}_{NB_PARTITIONS}", datastore2, OBJECT_TYPE, NB_PARTITIONS
    )


def test_checked_config_get_by_name(
    datastore: Datastore,
    datastore2: Datastore,
    config_id: int,
    config_id2: int,
    scrubber_db: ScrubberDb,
):
    assert datastore == scrubber_db.datastore_get(config_id)
    assert datastore2 == scrubber_db.datastore_get(config_id2)

    cfg2 = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
    cfg2prime = scrubber_db.config_add("cfg2", datastore2, ObjectType.SNAPSHOT, 42)
    cfg3 = scrubber_db.config_add("cfg3", datastore, ObjectType.SNAPSHOT, 43)

    assert scrubber_db.config_get_by_name("cfg2") == cfg2
    assert scrubber_db.config_get_by_name("cfg3") == cfg3

    # Check for unknown configuration
    assert scrubber_db.config_get_by_name("unknown config") is None

    # Check for duplicated configurations
    assert scrubber_db.config_get_by_name("cfg2", config_id) == cfg2
    assert scrubber_db.config_get_by_name("cfg2", config_id2) == cfg2prime


def test_datastore_get(datastore: Datastore, scrubber_db: ScrubberDb, config_id: int):
    assert scrubber_db.datastore_get(1) == datastore
    with pytest.raises(ValueError):
        scrubber_db.datastore_get(42)


def test_checked_partition_insert(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        part_gen = scrubber_db.checked_partition_iter_next(config_id)
        partition_id = next(part_gen)
    scrubber_db.checked_partition_upsert(config_id, partition_id, DATE + ONE_MINUTE)

    assert list(scrubber_db.checked_partition_iter(config_id)) == [
        (partition_id, NB_PARTITIONS, DATE, DATE + ONE_MINUTE)
    ]


def test_checked_partition_insert_two(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id = next(part_gen)
        scrubber_db.checked_partition_upsert(config_id, part_id, DATE + ONE_MINUTE)

        config_snp = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
        snp_part_gen = scrubber_db.checked_partition_iter_next(config_snp)
        snp_part_id = next(snp_part_gen)
        scrubber_db.checked_partition_upsert(config_snp, snp_part_id, DATE + ONE_MINUTE)

    assert set(scrubber_db.checked_partition_iter(config_id)) == {
        (part_id, NB_PARTITIONS, DATE, DATE + ONE_MINUTE),
    }
    assert set(scrubber_db.checked_partition_iter(config_snp)) == {
        (snp_part_id, 42, DATE, DATE + ONE_MINUTE),
    }


def test_checked_partition_get_next(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    config_snp = scrubber_db.config_add("cfg2", datastore, ObjectType.SNAPSHOT, 42)
    snp_part_gen = scrubber_db.checked_partition_iter_next(config_snp)
    dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)

    assert next(dir_part_gen) == 0
    assert next(snp_part_gen) == 0
    assert next(dir_part_gen) == 1
    assert next(dir_part_gen) == 2
    assert next(snp_part_gen) == 1
    assert next(snp_part_gen) == 2
    assert next(dir_part_gen) == 3

    date = datetime.datetime.now(tz=datetime.timezone.utc)
    scrubber_db.checked_partition_upsert(config_snp, 0, date)
    scrubber_db.checked_partition_upsert(config_id, 2, date)

    assert next(snp_part_gen) == 3
    assert next(dir_part_gen) == 4

    # iterate on all 64 possible partitions for the config config_id (5 of them
    # are already affected, so 59 to go)
    for i in range(59):
        assert next(dir_part_gen) == (i + 5)
    # we should be at the end of the partitions now...
    with pytest.raises(StopIteration):
        next(dir_part_gen)
    # check we still do not get anything after 63 with a new iterator
    assert list(scrubber_db.checked_partition_iter_next(config_id)) == []
    # check the database is OK with that
    with scrubber_db.transaction() as cur:
        cur.execute(
            "select max(partition_id) from checked_partition where config_id=%s",
            (config_id,),
        )
        assert cur.fetchone() == (63,)


def test_checked_partition_get_next_with_hole(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)

    # fill the checked_partition table
    list(zip(range(20), dir_part_gen))

    # one hole at a time
    for part_id in range(10):
        assert scrubber_db.checked_partition_reset(config_id, part_id)
        assert next(dir_part_gen) == part_id

    # a series of holes
    for part_id in range(0, 10, 2):
        assert scrubber_db.checked_partition_reset(config_id, part_id)

    for i in range(5):
        assert next(dir_part_gen) == 2 * i

    # all the holes are filled, next partition is 20
    assert next(dir_part_gen) == 20


def test_checked_partition_update(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id = next(dir_part_gen)
    scrubber_db.checked_partition_upsert(config_id, part_id, DATE + ONE_MINUTE)

    date2 = DATE + 2 * ONE_MINUTE
    scrubber_db.checked_partition_upsert(config_id, part_id, date2)

    assert list(scrubber_db.checked_partition_iter(config_id)) == [
        (part_id, NB_PARTITIONS, DATE, date2)
    ]

    date3 = DATE - ONE_MINUTE
    scrubber_db.checked_partition_upsert(config_id, part_id, date3)

    assert list(scrubber_db.checked_partition_iter(config_id)) == [
        (part_id, NB_PARTITIONS, DATE, date2)  # newest date wins
    ]


def test_checked_partition_get(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    with patch("swh.scrubber.db.now", return_value=DATE):
        dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id = next(dir_part_gen)
    assert scrubber_db.checked_partition_get_last_date(config_id, part_id) is None

    scrubber_db.checked_partition_upsert(config_id, part_id, DATE)

    assert scrubber_db.checked_partition_get_last_date(config_id, part_id) == DATE


def test_checked_partition_get_running(
    datastore: Datastore, scrubber_db: ScrubberDb, config_id: int
):
    assert list(scrubber_db.checked_partition_get_running(config_id)) == []
    with patch("swh.scrubber.db.now", return_value=DATE):
        dir_part_gen = scrubber_db.checked_partition_iter_next(config_id)
        part_id1 = next(dir_part_gen)
        part_id2 = next(dir_part_gen)
        part_id3 = next(dir_part_gen)

    assert scrubber_db.checked_partition_get_last_date(config_id, part_id1) is None
    assert scrubber_db.checked_partition_get_last_date(config_id, part_id2) is None
    assert scrubber_db.checked_partition_get_last_date(config_id, part_id3) is None

    assert list(scrubber_db.checked_partition_get_running(config_id)) == [
        (part_id1, DATE),
        (part_id2, DATE),
        (part_id3, DATE),
    ]

    scrubber_db.checked_partition_upsert(config_id, part_id2, DATE + ONE_MINUTE)
    assert list(scrubber_db.checked_partition_get_running(config_id)) == [
        (part_id1, DATE),
        (part_id3, DATE),
    ]

    scrubber_db.checked_partition_upsert(config_id, part_id1, DATE + ONE_MINUTE)
    assert list(scrubber_db.checked_partition_get_running(config_id)) == [
        (part_id3, DATE),
    ]

    scrubber_db.checked_partition_upsert(config_id, part_id3, DATE + ONE_MINUTE)
    assert list(scrubber_db.checked_partition_get_running(config_id)) == []



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Reads all objects in a swh-storage instance and recomputes their checksums."""

import json
import logging
from typing import Any, Dict, List

import attr

from swh.journal.client import get_journal_client
from swh.journal.serializers import kafka_to_value
from swh.model import model

from .base_checker import BaseChecker
from .db import Datastore, ScrubberDb

logger = logging.getLogger(__name__)


def get_datastore(journal_cfg) -> Datastore:
    if journal_cfg.get("cls") == "kafka":
        datastore = Datastore(
            package="journal",
            cls="kafka",
            instance=json.dumps(
                {
                    "brokers": journal_cfg["brokers"],
                    "group_id": journal_cfg["group_id"],
                    "prefix": journal_cfg["prefix"],
                }
            ),
        )
    else:
        raise NotImplementedError(
            f"JournalChecker(journal_client_config={journal_cfg!r}).datastore()"
        )
    return datastore


class JournalChecker(BaseChecker):
    """Reads a chunk of a swh-storage database, recomputes checksums, and
    reports errors in a separate database."""

    def __init__(
        self, db: ScrubberDb, config_id: int, journal_client_config: Dict[str, Any]
    ):
        super().__init__(db=db, config_id=config_id)

        if self.config.check_references:
            raise ValueError(
                "The journal checker cannot check for references, please set "
                "the 'check_references' to False in the config entry %s.",
                self.config_id,
            )
        self.journal_client_config = journal_client_config.copy()
        if "object_types" in self.journal_client_config:
            raise ValueError(
                "The journal_client configuration entry should not define the "
                "object_types field; this is handled by the scrubber configuration entry"
            )
        self.journal_client_config["object_types"] = [self.object_type.name.lower()]
        self.journal_client = get_journal_client(
            **self.journal_client_config,
            # Remove default deserializer; so process_kafka_values() gets the message
            # verbatim so it can archive it with as few modifications a possible.
            value_deserializer=lambda obj_type, msg: msg,
        )

    def run(self) -> None:
        """Runs a journal client with the given configuration.

        This method does not return, unless the client is configured with ``on_eof``
        parameter equals to ``EofBehavior.STOP`` (``stop`` in YAML).
        """
        self.journal_client.process(self.process_kafka_messages)

    def process_kafka_messages(self, all_messages: Dict[str, List[bytes]]):
        for object_type, messages in all_messages.items():
            logger.debug("Processing %s %s", len(messages), object_type)
            cls = getattr(model, object_type.capitalize())
            for message in messages:
                if object_type == "directory":
                    d = kafka_to_value(message)
                    (
                        has_duplicate_dir_entries,
                        object_,
                    ) = cls.from_possibly_duplicated_entries(
                        entries=tuple(
                            map(model.DirectoryEntry.from_dict, d["entries"])
                        ),
                        raw_manifest=d.get("raw_manifest"),
                    )
                    object_ = attr.evolve(object_, id=d["id"])
                    if has_duplicate_dir_entries:
                        self.statsd.increment(
                            "duplicate_directory_entries_total",
                            tags={"object_type": "directory"},
                        )
                else:
                    object_ = cls.from_dict(kafka_to_value(message))
                    has_duplicate_dir_entries = False
                real_id = object_.compute_hash()
                if object_.id != real_id or has_duplicate_dir_entries:
                    self.db.corrupt_object_add(object_.swhid(), self.config, message)
